@@ -153,6 +153,7 @@ sub html_startup {
 sub get_config {
 	my $use_cache = shift;
 	# usecache should match being in a cgi ($ENV{SCRIPT_NAME})
+	$htmlconfig = undef; # avoid double ram usage
 	if ($use_cache) {
 		$htmlconfig = generate_config($use_cache);
 	} else {
@@ -914,15 +915,15 @@ Munin-html creates the html pages.
 
 =head1 FILES
 
-	@@CONFDIR@@/munin.conf
-	@@DBDIR@@/datafile
-	@@LOGDIR@@/munin-html
-	@@HTMLDIR@@/*
-	@@STATEDIR@@/*
+	/etc/munin/munin.conf
+	/var/db/munin/datafile
+	/var/log/munin/munin-html
+	/var/www/htdocs/munin/*
+	/var/run/munin/*
 
 =head1 VERSION
 
-This is munin-html version @@VERSION@@
+This is munin-html version 2.0.18
 
 =head1 AUTHORS
 
