@@ -1,6 +1,6 @@
 package Munin::Master::Node;
 
-# $Id: Node.pm 4385 2011-09-16 23:30:47Z bldewolf $
+# $Id$
 
 # This module is used by UpdateWorker to keep in touch with a node and
 # parse some of the output.
@@ -14,9 +14,14 @@ use Munin::Common::Timeout;
 use Munin::Common::TLSClient;
 use Data::Dumper;
 use Log::Log4perl qw( :easy );
+use Time::HiRes qw( gettimeofday tv_interval );
+use IO::Socket::INET6;
 
 my $config = Munin::Master::Config->instance()->{config};
 
+# Quick version, to enable "DEBUG ... if $debug" constructs
+my $debug = $config->{debug};
+
 # Note: This timeout governs both small commands and waiting for the total
 # output of a plugin.  It is reset for each read.
 
@@ -28,8 +33,10 @@ sub new {
         port    => $port,
         host    => $host,
         tls     => undef,
-        socket  => undef,
-        master_capabilities => qw(multigraph),
+        reader  => undef,
+        pid     => undef,
+        writer  => undef,
+        master_capabilities => "multigraph dirtyconfig",
         io_timeout => 120,
 	configref => $configref,
     };
@@ -43,9 +50,9 @@ sub do_in_session {
 
     if ($self->_do_connect()) {
 	$self->_run_starttls_if_required();
-	$block->();
+	my $exit_value = $block->();
 	$self->_do_close();
-	return 1; # If we're still here
+	return { exit_value => $exit_value }; # If we're still here
     }
     return 0;  # _do_connect failed.
 }
@@ -58,34 +65,90 @@ sub _do_connect {
     LOGCROAK("[FATAL] No address!  Did you forget to set 'update no' or to set 'address <IP>' ?")
 	if !defined($self->{address});
 
-    if (! ( $self->{socket} = IO::Socket::INET->new(
-		PeerAddr  => $self->{address},
-		PeerPort  => $self->{port},
+    # Check if it's an URI or a plain host
+    use URI;
+
+    # Parameters are space-separated from the main address
+    my ($url, $params) = split(/ +/, $self->{address}, 2);
+    my $uri = new URI($url);
+
+    # If address is only "ssh://host/" $params will not get set
+    $params = "" unless defined $params;
+
+    # If the scheme is not defined, it's a plain host. 
+    # Prefix it with munin:// to be able to parse it like others
+    $uri = new URI("munin://" . $url) unless $uri->scheme;
+    LOGCROAK("[FATAL] '$url' is not a valid address!") unless $uri->scheme;
+
+    if ($uri->scheme eq "munin") {
+        $self->{reader} = $self->{writer} = IO::Socket::INET6->new(
+		PeerAddr  => $uri->host,
+		PeerPort  => $self->{port} || 4949,
 		LocalAddr => $config->{local_address},
 		Proto     => 'tcp', 
-		Timeout   => $config->{timeout}) ) ) {
-	ERROR "Failed to connect to node $self->{address}:$self->{port}/tcp : $!";
-	return 0;
+		MultiHomed => 1,
+		Timeout   => $config->{timeout}
+	);
+	if (! $self->{reader} ) {
+		ERROR "Failed to connect to node $self->{address}:$self->{port}/tcp : $!";
+		return 0;
+	}
+    } elsif ($uri->scheme eq "ssh") {
+	    my $ssh_command = "ssh -o ChallengeResponseAuthentication=no -o StrictHostKeyChecking=no ";
+	    my $user_part = ($uri->user) ? ($uri->user . "@") : "";
+	    my $remote_cmd = ($uri->path ne '/') ? $uri->path : "";
+
+	    # Add any parameter to the cmd
+	    my $remote_connection_cmd = $ssh_command . " -p " . $uri->port . " " . $user_part . $uri->host . " " . $remote_cmd . " " . $params;
+
+	    # Open a triple pipe
+   	    use IPC::Open3;
+
+	    $self->{reader} = new IO::Handle();
+	    $self->{writer} = new IO::Handle();
+	    $self->{stderr} = new IO::Handle();
+
+	    DEBUG "[DEBUG] open3($remote_connection_cmd)";
+	    $self->{pid} = open3($self->{writer}, $self->{reader}, $self->{stderr}, $remote_connection_cmd);
+            ERROR "Failed to connect to node $self->{address} : $!" unless $self->{pid};
+    } elsif ($uri->scheme eq "cmd") {
+        # local commands should ignore the username, url and host
+        my $local_cmd = $uri->path;
+        my $local_pipe_cmd = "$local_cmd $params";
+
+	    # Open a triple pipe
+   	    use IPC::Open3;
+
+	    $self->{reader} = new IO::Handle();
+	    $self->{writer} = new IO::Handle();
+	    $self->{stderr} = new IO::Handle();
+
+	    DEBUG "[DEBUG] open3($local_pipe_cmd)";
+	    $self->{pid} = open3($self->{writer}, $self->{reader}, $self->{stderr}, $local_pipe_cmd);
+            ERROR "Failed to execute local command: $!" unless $self->{pid};
+    } else {
+	    ERROR "Unknown scheme : " . $uri->scheme;
+	    return 0;
     }
 
-    my $greeting = $self->_node_read_single();
-    $self->{node_name} = $self->_extract_name_from_greeting($greeting);
+    # check all the lines until we find one that matches the expected
+    # greeting; ignore anything that doesn't look like it as long as
+    # there is output. This allows to accept SSH connections where
+    # lastlog or motd is used.
+    until(defined($self->{node_name})) {
+	my $greeting = $self->_node_read_single();
+	if (!$greeting) {
+	    die "[ERROR] Got unknown reply from node ".$self->{host}."\n";
+	}
+
+	if ($greeting =~ /\#.*(?:lrrd|munin) (?:client|node) at (\S+)/i) {
+	    $self->{node_name} = $1;
+	}
+    };
+
     return 1;
 }
 
-
-sub _extract_name_from_greeting {
-    my ($self, $greeting) = @_;
-    if (!$greeting) {
-	die "[ERROR] Got no reply from node ".$self->{host}."\n";
-    }
-    if ($greeting !~ /\#.*(?:lrrd|munin) (?:client|node) at (\S+)/) {
-	die "[ERROR] Got unknown reply from node ".$self->{host}."\n";
-    }
-    return $1;
-}
-
-
 sub _run_starttls_if_required {
     my ($self) = @_;
 
@@ -99,7 +162,7 @@ sub _run_starttls_if_required {
     $self->{tls} = Munin::Common::TLSClient->new({
         DEBUG        => $config->{debug},
         logger       => sub { $logger->warn(@_) },
-        read_fd      => fileno($self->{socket}),
+        read_fd      => fileno($self->{reader}),
         read_func    => sub { _node_read_single($self) },
         tls_ca_cert  => $config->{tls_ca_certificate},
         tls_cert     => $config->{tls_certificate},
@@ -108,7 +171,7 @@ sub _run_starttls_if_required {
         tls_vdepth   => $config->{tls_verify_depth},
         tls_verify   => $config->{tls_verify_certificate},
         tls_match    => $config->{tls_match},
-        write_fd     => fileno($self->{socket}),
+        write_fd     => fileno($self->{writer}),
         write_func   => sub { _node_write_single($self, @_) },
     });
 
@@ -124,8 +187,17 @@ sub _run_starttls_if_required {
 sub _do_close {
     my ($self) = @_;
 
-    close $self->{socket};
-    $self->{socket} = undef;
+    close $self->{reader};
+    close $self->{writer};
+    $self->{reader} = undef;
+    $self->{writer} = undef;
+
+    # Close stderr if needed
+    close $self->{stderr} if $self->{stderr};
+    $self->{stderr} = undef if $self->{stderr};
+
+    # Reap the underlying process
+    waitpid($self->{pid}, 0) if (defined $self->{pid});
 }
 
 
@@ -181,7 +253,7 @@ sub list_plugins {
 
 
 sub parse_service_config {
-    my ($self, $service, @lines) = @_;
+    my ($self, $service, $lines) = @_;
 
     my $errors;
     my $correct;
@@ -220,9 +292,12 @@ sub parse_service_config {
 
     new_service($service);
 
-    for my $line (@lines) {
+    # every 'N' has the same value. Should not take parsing time into the equation
+    my $now = time;
 
-	DEBUG "[CONFIG from $plugin] $line";
+    for my $line (@$lines) {
+
+	DEBUG "[CONFIG from $plugin] $line" if $debug;
 
 	if ($line =~ /\# timeout/) {
 	    die "[ERROR] Timeout error on $nodedesignation during fetch of $plugin. \n";
@@ -232,34 +307,67 @@ sub parse_service_config {
         next if $line =~ /^\#/;
 
 	if ($line =~ m{\A multigraph \s+ (.+) }xms) {
-	    $correct++;
-
 	    push_graphorder($service);
 
 	    $service = $1;
 
 	    if ($service eq 'multigraph') {
-		die "[ERROR] SERVICE can't be named \"$service\" in plugin $plugin on ".$self->{host}."/".$self->{address}."/".$self->{port};
+		ERROR "[ERROR] SERVICE can't be named \"$service\" in plugin $plugin on ".$self->{host}."/".$self->{address}."/".$self->{port};
+                $errors++;
+                last;
 	    }
-	    new_service($service);
+            if ($service =~ /(^\.|\.$|\.\.)/) {
+                ERROR "[ERROR] SERVICE \"$service\" contains dots in wrong places in plugin $plugin on ".$self->{host}."/".$self->{address}."/".$self->{port};
+                $errors++;
+                last;
+            }
+            if ($service !~ m/^[-\w.:.]+$/) {
+                ERROR "[ERROR] SERVICE \"$service\" contains weird characters in plugin $plugin on ".$self->{host}."/".$self->{address}."/".$self->{port};
+                $errors++;
+                last;
+            }
+	    new_service($service) unless $global_config->{$service};
 	    DEBUG "[CONFIG multigraph $plugin] Service is now $service";
+	    $correct++;
 	}
-	elsif ($line =~ m{\A ([^\s\.]+) \s+ (.+) }xms) {
+	elsif ($line =~ m{\A ([^\s\.]+) \s+ (.+?) \s* $}xms) {
 	    $correct++;
 
 	    my $label = $self->_sanitise_fieldname($1);
 
-            push @{$global_config->{$service}}, [$label, $2];
-            DEBUG "[CONFIG graph global $plugin] $service->$label = $2";
+	    # add to config if not already here
+	    push @{$global_config->{$service}}, [$label, $2]
+	    	unless grep { $_->[0] eq $label }  @{$global_config->{$service}};
+            DEBUG "[CONFIG graph global $plugin] $service->$label = $2" if $debug;
+        } elsif ($line =~ m{\A ([^\.]+)\.value \s+ (.+?) \s* $}xms) {
+	    $correct++;
+	    # Special case for dirtyconfig
+            my ($ds_name, $value, $when) = ($1, $2, $now);
+            
+	    $ds_name = $self->_sanitise_fieldname($ds_name);
+	    if ($value =~ /^(\d+):(.+)$/) {
+		$when = $1;
+		$value = $2;
+	    }
+            DEBUG "[CONFIG dirtyconfig $plugin] Storing $value from $when in $ds_name";
+
+	    # Creating the datastructure if not created already
+            $data_source_config->{$service}{$ds_name} ||= {};
+            $data_source_config->{$service}{$ds_name}{when} ||= [];
+            $data_source_config->{$service}{$ds_name}{value} ||= [];
+	
+	    # Saving the timed value in the datastructure
+	    push @{$data_source_config->{$service}{$ds_name}{when}}, $when;
+	    push @{$data_source_config->{$service}{$ds_name}{value}}, $value;
         }
-	elsif ($line =~ m{\A ([^\.]+)\.([^\s]+) \s+ (.+) }xms) {
+	elsif ($line =~ m{\A ([^\.]+)\.([^\s]+) \s+ (.+?) \s* $}xms) {
 	    $correct++;
 	    
             my ($ds_name, $ds_var, $ds_val) = ($1, $2, $3);
             $ds_name = $self->_sanitise_fieldname($ds_name);
             $data_source_config->{$service}{$ds_name} ||= {};
             $data_source_config->{$service}{$ds_name}{$ds_var} = $ds_val;
-            DEBUG "[CONFIG dataseries $plugin] $service->$ds_name.$ds_var = $ds_val";
+            DEBUG "[CONFIG dataseries $plugin] $service->$ds_name.$ds_var = $ds_val" if $debug;
             push ( @graph_order, $ds_name ) if $ds_var eq 'label';
         }
 	else {
@@ -283,17 +391,37 @@ sub parse_service_config {
 sub fetch_service_config {
     my ($self, $service) = @_;
 
+    my $t0 = [gettimeofday];
+
     DEBUG "[DEBUG] Fetching service configuration for '$service'";
     $self->_node_write_single("config $service\n");
 
     # The whole config in one fell swoop.
-    my @lines = $self->_node_read();
+    my $lines = $self->_node_read();
+
+    my $elapsed = tv_interval($t0);
+
+    my $nodedesignation = $self->{host}."/".$self->{address}."/".$self->{port};
+    DEBUG "[DEBUG] config: $elapsed sec for '$service' on $nodedesignation";
 
     $service = $self->_sanitise_plugin_name($service);
 
-    return $self->parse_service_config($service,@lines);
+    return $self->parse_service_config($service, $lines);
 }
 
+sub spoolfetch {
+    my ($self, $timestamp) = @_;
+
+    DEBUG "[DEBUG] Fetching spooled services since $timestamp (" . localtime($timestamp) . ")";
+    $self->_node_write_single("spoolfetch $timestamp\n");
+
+    # The whole stuff in one fell swoop.
+    my $lines = $self->_node_read();
+
+    # using the multigraph parsing. 
+    # Using "__root__" as a special plugin name. 
+    return $self->parse_service_config("__root__", $lines);
+}
 
 sub _validate_data_sources {
     my ($self, $all_data_source_config) = @_;
@@ -315,7 +443,7 @@ sub _validate_data_sources {
 
 
 sub parse_service_data {
-    my ($self, $service, @lines) = @_;
+    my ($self, $service, $lines) = @_;
 
     my $plugin = $service;
     my $errors = 0;
@@ -330,7 +458,10 @@ sub parse_service_data {
     DEBUG "[DEBUG] Now parsing fetch output from plugin $plugin on ".
 	$nodedesignation;
 
-    for my $line (@lines) {
+    # every 'N' has the same value. Should not take parsing time into the equation
+    my $now = time;
+
+    for my $line (@$lines) {
 
 	DEBUG "[FETCH from $plugin] $line";
 
@@ -343,19 +474,29 @@ sub parse_service_data {
         next if $line =~ /^\#/;
 
 	if ($line =~ m{\A multigraph \s+ (.+) }xms) {
-	    $correct++;
-
 	    $service = $1;
+            if ($service =~ /(^\.|\.$|\.\.)/) {
+                ERROR "[ERROR] SERVICE \"$service\" contains dots in wrong places in plugin $plugin on ".$self->{host}."/".$self->{address}."/".$self->{port};
+                $errors++;
+                last;
+            }
+            if ($service !~ m/^[-\w.:.]+$/) {
+                ERROR "[ERROR] SERVICE \"$service\" contains weird characters in plugin $plugin on ".$self->{host}."/".$self->{address}."/".$self->{port};
+                $errors++;
+                last;
+            }
 	    $values{$service} = {};
 
 	    if ($service eq 'multigraph') {
+                $errors++;
 		ERROR "[ERROR] SERVICE can't be named \"$service\" in plugin $plugin on ".
 		    $nodedesignation;
-		croak("Plugin error.  Please consult the log.");
+                last;
 	    }
+	    $correct++;
 	}
 	elsif ($line =~ m{\A ([^\.]+)\.value \s+ ([\S:]+) }xms) {
-            my ($data_source, $value, $when) = ($1, $2, 'N');
+            my ($data_source, $value, $when) = ($1, $2, $now);
 
 	    $correct++;
 
@@ -368,12 +509,12 @@ sub parse_service_data {
 		$value = $2;
 	    }
 
-	    $values{$service}{$data_source} ||= {};
+	    $values{$service}{$data_source} ||= { when => [], value => [], };
 
-            $values{$service}{$data_source}{value} = $value;
-            $values{$service}{$data_source}{when}  = $when;
+	    push @{$values{$service}{$data_source}{when}}, $when;
+	    push @{$values{$service}{$data_source}{value}}, $value;
         }
-	elsif ($line =~ m{\A ([^\.]+)\.extinfo \s+ (.+) }xms) {
+	elsif ($line =~ m{\A ([^\.]+)\.extinfo \s+ (.+?) \s* $}xms) {
 	    # Extinfo is used in munin-limits
             my ($data_source, $value) = ($1, $2);
 	    
@@ -393,7 +534,9 @@ sub parse_service_data {
         }
     }
     if ($errors) {
-	WARN "[WARNING] $errors lines had errors while $correct lines were correct in data from 'fetch $plugin' on $nodedesignation";
+	my $percent = ($errors / ($errors + $correct)) * 100; 
+	$percent = sprintf("%.2f", $percent);
+	WARN "[WARNING] $errors lines had errors while $correct lines were correct ($percent%) in data from 'fetch $plugin' on $nodedesignation";
     }
 
     return %values;
@@ -403,13 +546,31 @@ sub parse_service_data {
 sub fetch_service_data {
     my ($self, $plugin) = @_;
 
+    my $t0 = [gettimeofday];
+
     $self->_node_write_single("fetch $plugin\n");
 
-    my @lines = $self->_node_read();
+    my $lines = $self->_node_read_fast();
+    
+    my $elapsed = tv_interval($t0);
+    my $nodedesignation = $self->{host}."/".$self->{address}."/".$self->{port};
+    DEBUG "[DEBUG] data: $elapsed sec for '$plugin' on $nodedesignation";
 
     $plugin = $self->_sanitise_plugin_name($plugin);
 
-    return $self->parse_service_data($plugin,@lines);
+    return $self->parse_service_data($plugin, $lines);
+}
+
+sub quit {
+    my ($self) = @_;
+
+    my $t0 = [gettimeofday];
+    $self->_node_write_single("quit \n");
+    my $elapsed = tv_interval($t0);
+    my $nodedesignation = $self->{host}."/".$self->{address}."/".$self->{port};
+    DEBUG "[DEBUG] quit: $elapsed sec on $nodedesignation";
+
+    return 1;
 }
 
 
@@ -423,7 +584,7 @@ sub _sanitise_plugin_name {
 
 
 sub _sanitise_fieldname {
-    # http://munin.projects.linpro.no/wiki/notes_on_datasource_names
+    # http://munin-monitoring.org/wiki/notes_on_datasource_names
     my ($self, $name) = @_;
 
     $name =~ s/^[^A-Za-z_]/_/;
@@ -439,12 +600,11 @@ sub _node_write_single {
     DEBUG "[DEBUG] Writing to socket: \"$text\".";
     my $timed_out = !do_with_timeout($self->{io_timeout}, sub {
         if ($self->{tls} && $self->{tls}->session_started()) {
-            $self->{tls}->write($text)
-                or exit 9;
-        }
-        else {
-            print { $self->{socket} } $text;
+            $self->{tls}->write($text) or exit 9;
+        } else {
+            print { $self->{writer} } $text;
         }
+	return 1;
     });
     if ($timed_out) {
         LOGCROAK "[FATAL] Socket write timed out to ".$self->{host}.
@@ -463,9 +623,11 @@ sub _node_read_single {
           $res = $self->{tls}->read();
       }
       else {
-          $res = readline $self->{socket};
+          $res = readline $self->{reader};
       }
-      chomp $res if defined $res;
+      # Remove \r *and* \n. Normally only one, since we read line per line.
+      $res =~ tr/\x{d}\x{a}//d if defined $res;
+      return 1;
     });
     if ($timed_out) {
         LOGCROAK "[FATAL] Socket read timed out to ".$self->{host}.
@@ -476,10 +638,41 @@ sub _node_read_single {
 	# aren't we supposed to be in "do in session"?
 	LOGCROAK "[FATAL] Socket read from ".$self->{host}." failed.  Terminating process.";
     }
-    DEBUG "[DEBUG] Reading from socket to ".$self->{host}.": \"$res\".";
+    DEBUG "[DEBUG] Reading from socket to ".$self->{host}.": \"$res\"." if $debug;
     return $res;
 }
 
+sub _node_read_fast {
+	my ($self) = @_;
+
+	# We cannot bypass the IO if using TLS
+	# so just reverting to normal mode.
+	return _node_read(@_) if $self->{tls};
+
+	# Disable Buffering here, to be able to use sysread()
+	local $| = 1;
+
+	my $io_src = $self->{reader};
+        my $buf;
+    	my $offset = 0;
+        while (my $read_len = sysread($io_src, $buf, 4096, $offset)) {
+		$offset += $read_len;
+
+		# Stop when we read a \n.\n
+		# ... No need to have a full regex : simple index()
+		my $start_offset = $offset - $read_len - 3;
+		$start_offset = 0 if $start_offset < 0;
+		last if index($buf, "\n.\n", $start_offset) >= 0;
+
+		# if empty, the client only sends a plain ".\n"
+		last if $buf eq ".\n";
+        }
+
+	# Remove the last line that only contains ".\n"
+	$buf =~ s/\.\n$//;
+
+	return [ split(/\n/, $buf) ];
+}
 
 sub _node_read {
     my ($self) = @_;
@@ -492,9 +685,20 @@ sub _node_read {
     }
 
     DEBUG "[DEBUG] Reading from socket: \"".(join ("\\n",@array))."\".";
-    return @array;
+    return \@array;
 }
 
+# Defines the URL::scheme for munin
+package URI::munin;
+
+# We are like a generic server
+require URI::_server;
+@URI::munin::ISA=qw(URI::_server);
+
+# munin://HOST[:PORT]
+
+sub default_port { return 4949; }
+
 1;
 
 __END__
@@ -540,3 +744,4 @@ FIX
 FIX
 
 =back
+
